/*
 * Copyright (c) 1995, 1996, 1997, 1999 The University of Utah and
 * the Computer Systems Laboratory at the University of Utah (CSL).
 *
 * This file is part of Flick, the Flexible IDL Compiler Kit.
 *
 * Flick is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Flick is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Flick; see the file COPYING.  If not, write to
 * the Free Software Foundation, 59 Temple Place #330, Boston, MA 02111, USA.
 */

#include <assert.h>
#include <mom/libmint.h>
#include <mom/c/libcast.h>
#include "khazana.h"

/*
 * Return statements that specify whether or not we need to do bit translation.
 * I.E. in Khazana, this would return
 *   for decoding same endianness, #if 0, #else, and #endif
 *   for decoding opposite endianness, #if 1, #else, and #endif
 *   for encoding, we never have to swap
 * However, single chars never need byte swapped.
 * 
 * The parameter specifies whether to output the if, else or endif
 * (0, 1, 2 respectively)
 */

static char *decode_if_text =
	"#if %d /* are we in the swapping decode or non-swapping decode */";
static char *decode_else_text = "#else";
static char *decode_endif_text = "#endif";

static char *encode_if_text = "#if 0 /* We never byte swap on encoding */";
static char *encode_else_text = "#else /* Always bcopy */";
static char *encode_endif_text = "#endif /* alway bcopy */";

static char *never_text = "#if 0 /* No need to swap this */";

cast_stmt khazana_mu_state::mu_bit_translation_necessary(int which,
							 mint_ref itype)
{
	cast_stmt macro = cast_new_stmt(CAST_STMT_TEXT);
	int decode = op & MUST_DECODE;
	int size, align;
	char *name;
	
	/* First, determine the size of the type in question */
	get_prim_params(itype, &size, &align, &name);
	
	/* Now determine which statement to return */
	switch (which) {
	case 0: // #if
		if (size == 1)
			/* Never have to swap single bytes! */
			macro->cast_stmt_u_u.text = never_text;
		else
			macro->cast_stmt_u_u.text
				= decode
				? flick_asprintf(decode_if_text, should_swap)
				: encode_if_text;
		break;
	case 1: // #else
		macro->cast_stmt_u_u.text = decode
					    ? decode_else_text
					    : encode_else_text;
		break;
	case 2: // #endif
		macro->cast_stmt_u_u.text = decode
					    ? decode_endif_text
					    : encode_endif_text;
		break;
	default:
		panic("Invalid value passed to "
		      "mu_bit_translation_necessary\n");
	}
	return macro;
}

/* End of file */
