/*
 * Copyright (c) 1995, 1996, 1997, 1998, 1999 The University of Utah and
 * the Computer Systems Laboratory at the University of Utah (CSL).
 *
 * This file is part of Flick, the Flexible IDL Compiler Kit.
 *
 * Flick is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Flick is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Flick; see the file COPYING.  If not, write to
 * the Free Software Foundation, 59 Temple Place #330, Boston, MA 02111, USA.
 */

#include <assert.h>
#include <string.h>
#include <mom/libmint.h>
#include <mom/c/libcast.h>
#include <mom/c/libpres_c.h>
#include "trapeze.h"

void init_be(be_flags /*the_flags*/)
{
	cast_language = CAST_C;
}

trapeze_protocol_t what_protocol(pres_c_1 *pres)
{
	if (strcmp(pres->pres_context, "sun") == 0)
		return TRAPEZE_ONC;
	return TRAPEZE;
}

trapeze_mu_state::trapeze_mu_state(pres_c_1 *mu_pres,
				   mu_state_op mu_op,
				   int mu_assumptions,
				   char *mu_which,
				   int swap)
	: mem_mu_state(mu_pres, mu_op, mu_assumptions, 2, 0, 120, mu_which)
{
	should_swap = swap;
	has_payload = 0;
	protocol = what_protocol(mu_pres);
}

trapeze_mu_state::trapeze_mu_state(const trapeze_mu_state &must)
	: mem_mu_state(must)
{
	should_swap = must.should_swap;
	has_payload = must.has_payload;
	protocol = must.protocol;
}

mu_state *trapeze_mu_state::another(mu_state_op mu_op)
{
	return new trapeze_mu_state(pres, mu_op, assumptions, get_which_stub(),
				    should_swap);
}

mu_state *trapeze_mu_state::clone()
{
	return new trapeze_mu_state(*this);
}

trapeze_target_mu_state::trapeze_target_mu_state(pres_c_1 *mu_pres,
						 mu_state_op mu_op,
						 int mu_assumptions,
						 char *mu_which)
	: target_mu_state(mu_pres, mu_op, mu_assumptions, mu_which)
{
	protocol = what_protocol(mu_pres);
}

trapeze_target_mu_state::trapeze_target_mu_state(
	const trapeze_target_mu_state &must)
	: target_mu_state(must)
{
	protocol = must.protocol;
}

mu_state *trapeze_target_mu_state::clone()
{
	return new trapeze_target_mu_state(*this);
}

char *trapeze_target_mu_state::get_encode_name() 
{
	if (protocol == TRAPEZE_ONC)
		return "xdr";
	return "cdr";
}

char *trapeze_target_mu_state::get_be_name()
{
	return "trapeze";
}

trapeze_client_mu_state::trapeze_client_mu_state(pres_c_1 *mu_pres,
						 mu_state_op mu_op,
						 int mu_assumptions,
						 char *mu_which)
	: client_mu_state(mu_pres, mu_op, mu_assumptions, mu_which)
{
}

trapeze_client_mu_state::trapeze_client_mu_state(const trapeze_client_mu_state &must)
	: client_mu_state(must)
{
}

mu_state *trapeze_client_mu_state::clone()
{
	return new trapeze_client_mu_state(*this);
}

char *trapeze_client_mu_state::get_encode_name() 
{
	return "cdr";
}

char *trapeze_client_mu_state::get_be_name()
{
    return "trapeze";
}

char *trapeze_mu_state::get_encode_name() 
{
	if (protocol == TRAPEZE_ONC) {
		if (should_swap == TRAPEZE_NO_SWAP || !(op & MUST_DECODE))
			return "xdr";
		else
			return "xdr_swap";
	} else {
		if (should_swap == TRAPEZE_NO_SWAP || !(op & MUST_DECODE))
			return "cdr";
		else
			return "cdr_swap";
	}
}

char *trapeze_mu_state::get_be_name()
{
   return "trapeze";
}

void trapeze_mu_state::mu_server_func(pres_c_inline inl, mint_ref tn_r,
				      pres_c_server_func *sfunc,
				      pres_c_skel *sskel)
{
	has_payload = 0; /* reset the payload ptr */
	mem_mu_state::mu_server_func(inl, tn_r, sfunc, sskel);
}
	
/* This handles PRES_C_MAPPING_SYSTEM_EXCEPTION nodes,
   which are used to opaquely represent system exceptions.
   They're opaque, because their presentation varies greatly.
*/
void trapeze_mu_state::mu_mapping_system_exception(cast_expr expr,
						   cast_type /* ctype */,
						   mint_ref itype)
{
	mint_def *idef = &pres->mint.defs.defs_val[itype];
	
	assert(idef->kind == MINT_SYSTEM_EXCEPTION);
	
	// This is a system exception - deal with it separately
	// Not sure if we should break the chunk &/| glob...
	break_chunk();
	cast_expr macro = cast_new_expr_name(
		flick_asprintf("flick_%s_%s_system_exception",
			       pres->pres_context,
			       get_buf_name()));
	
	cast_expr cex = cast_new_expr_call_3(
		macro,
		expr,
		cast_new_expr_name(get_encode_name()),
		cast_new_expr_name(get_be_name()));
	
	add_stmt(cast_new_stmt_expr(cex));
}

void
w_header_includes(pres_c_1 *pres)
{
	w_printf("#define FLICK_PROTOCOL_TRAPEZE%s\n",
		 (what_protocol(pres) == TRAPEZE_ONC)? "_ONC" : "");
	w_printf("#include <flick/link/trapeze.h>\n");
	w_printf("#include <flick/encode/%cdr.h>\n",
		 (what_protocol(pres) == TRAPEZE_ONC)? 'x' : 'c');
}

/*
 * Replace the operation identifier with a sequentially numbered
 * constant value.  This reduces the overhead of [un]marshaling an
 * unbounded array for the operation name.
 */
void replace_operation_ids(pres_c_1 *pres,
			   mint_ref itype,
			   pres_c_inline inl)
{
	static int counter = 0;
	
	assert(pres->mint.defs.defs_val[itype].kind == MINT_UNION);
	mint_union_def *udef = &(pres->mint.defs.defs_val[itype].mint_def_u.
				 union_def);
	
	assert(inl->kind == PRES_C_INLINE_COLLAPSED_UNION);
	pres_c_inline_collapsed_union *cuinl
		= &(inl->pres_c_inline_u_u.collapsed_union);
	
	/* Find the appropriate case. */
	mint_ref case_var_r;
	for (int case_num = 0; ; case_num++) {
		if (case_num >= (signed int) udef->cases.cases_len) {
			if (udef->dfault == mint_ref_null)
				panic("union collapsed on nonexistent case");
			case_var_r = udef->dfault;
			break;
		}

		/* replace the mint_const here with a manufactured
		   constant integer, so we don't have to [un]marshal
		   an unbounded array -- just a 4-byte int! */
		if (mint_const_cmp(udef->cases.cases_val[case_num].val,
				   cuinl->discrim_val) == 0) {
			if (cuinl->discrim_val->kind != MINT_CONST_INT) {
				mint_const myconst =
					mint_new_const_int(counter++);
				cuinl->discrim_val = myconst;
				udef->cases.cases_val[case_num].val = myconst;
				udef->discrim =
					pres->mint.standard_refs.signed32_ref;
			}
			break;
		}
	}
}

/* End of file. */
