/*
 * Copyright (c) 1998, 1999 The University of Utah and
 * the Computer Systems Laboratory at the University of Utah (CSL).
 *
 * This file is part of Flick, the Flexible IDL Compiler Kit.
 *
 * Flick is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Flick is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Flick; see the file COPYING.  If not, write to
 * the Free Software Foundation, 59 Temple Place #330, Boston, MA 02111, USA.
 */

#include <mom/compiler.h>
#include <mom/c/pfe.hh>
#include <mom/c/libcast.h>
#include <mom/c/libpres_c.h>

#include <mom/c/pg_corbaxx.hh>

/*
 * If you override this method, you should also override `p_interface_type',
 * which handles normal interface declarations.
 */
void pg_corbaxx::p_forward_type(p_type_collection **out_ptc)
{
	p_type_collection *ptc;
	p_type_node *ptn;
	cast_scoped_name scn;
	cast_type ctype;
	pres_c_mapping map;
	tag_list *tl;
	tag_data *pt_td;
	union tag_data_u data;
	char *int_name;
	
	int_name = calc_name_from_ref(cur_aoi_idx);
	scn = cast_new_scoped_name(int_name, 0);
	
	map = pres_c_new_mapping(PRES_C_MAPPING_REFERENCE);
	map->pres_c_mapping_u_u.ref.kind = PRES_C_REFERENCE_COPY;
	map->pres_c_mapping_u_u.ref.ref_count = 1;
	
	ctype = cast_new_type_scoped_name(scn);
	
	ptc = p_new_type_collection("");
	
	if( *out_ptc )
		(*out_ptc)->set_collection_ref(ptc);
	else
		*out_ptc = ptc;
	
	/* Construct the type collection for the interface */
	ptc = *out_ptc;
	pt_td = &(find_tag(out_pres->pres_attrs, "pres_type")->data);
	tl = create_tag_list(0);
	data.tl = tl;
	ptc->set_attr_index(append_tag_data(pt_td, data));
	ptc->set_tag_list(tl);
	add_tag(tl, "name", TAG_STRING, ptc->get_name());
	add_tag(tl, "idl_type", TAG_STRING, "interface");
	add_tag(tl, "size", TAG_STRING, "variable");
	
	ptn = new p_type_node;
	ptn->set_name("forward_declaration");
	ptn->set_format(0);
	ctype = cast_new_type(CAST_TYPE_CLASS_NAME);
	ctype->cast_type_u_u.class_name = scn;
	ptn->set_type(ctype);
	ptc->add_type("default", ptn);
	
	ptn = new p_type_node;
	ptn->set_name("pointer");
	ptn->set_format("%s_ptr");
	ctype = cast_new_type_scoped_name(scn);
	ptn->set_type(cast_new_pointer_type(ctype));
	ptn->set_mapping(map);
	ptc->add_type("default", ptn);
	
	ptn = new p_type_node;
	ptn->set_name("definition");
	ptn->set_format("%s");
	ptn->set_type(ctype);
	ptn->set_mapping(map);
	ptc->add_type("default", ptn);
	
	ptc->set_id(get_repository_id(cur_aoi_idx));
	
	p_type_var(ptc, a(cur_aoi_idx).binding);
	
	p_out_class(ptc, a(cur_aoi_idx).binding);
	
	cast_scope dummy_scope = cast_new_scope(0);
	aoi_type at;
	
	at = a(cur_aoi_idx).binding;
	inclusion included = IMPLIED;
	
	/* If a forward interface has no definition we still
	   need to generate its interface functions for
	   reference in the back end */
	if( (at->kind == AOI_FWD_INTRFC) &&
	    (at->aoi_type_u_u.fwd_intrfc_def == -1) ) {
		a(cur_aoi_idx).included = IMPLIED;
		cast_add_scope_name(&current_scope_name,
				    name,
				    null_template_arg_array);
		p_interface_funcs(ptc, &dummy_scope);
		cast_del_scope_name(&current_scope_name);
	}
	
	if( gen_client ) {
		/* We're in the client so we need to squelch the poa stuff */
		included = a(cur_aoi_idx).included;
		a(cur_aoi_idx).included = IMPLIED;
	}
	if( gen_server &&
	    (a(cur_aoi_idx).included == SUPPRESSED) ) {
		/* We're in the server so we need to let the poa stuff out */
		a(cur_aoi_idx).included = NOT_INCLUDED;
	}
	p_poa_class(ptc);
	p_poa_tie(ptc);
	if( gen_client )
		a(cur_aoi_idx).included = included;
	
	/* Set the format for the definition to "" so that it won't
	   be output, we just need to write out a forward class
	   definition.  If the interface is defined then it will
	   replace these definition's */
	ptc->find_type("definition")->set_format("");
	ptc->get_collection_ref()->find_type("definition")->set_format("");
}

/* End of file. */

