/*
 * Copyright (c) 1997, 1998 The University of Utah and
 * the Computer Systems Laboratory at the University of Utah (CSL).
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software is hereby
 * granted provided that (1) source code retains these copyright, permission,
 * and disclaimer notices, and (2) redistributions including binaries
 * reproduce the notices in supporting documentation.
 *
 * THE UNIVERSITY OF UTAH AND CSL ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS
 * IS" CONDITION.  THE UNIVERSITY OF UTAH AND CSL DISCLAIM ANY LIABILITY OF
 * ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * CSL requests users of this software to return to csl-dist@cs.utah.edu any
 * improvements that they make and grant CSL redistribution rights.
 */

#ifndef __flick_pres_all_h
#define __flick_pres_all_h

/* This include is only for memset() when outputted by
   cast_new_expr_assign_to_zero() to initialize the _return value to 0 in the
   case of arrays/structs/unions/named types */
#include <string.h>

/* Dough-head error reporting */
/* These 2 includes are only for the dough-header error reporting... */
#include <stdio.h>
#include <unistd.h>
#define flick_error(_pres, _phase, _stub) {			\
	printf("ERROR Process #%d, Line #%d, File '%s'!\n",	\
	       (int)getpid(), __LINE__, __FILE__);		\
	printf("Presentation:\t%s\nPhase:\t\t%s\nFrom:\t\t%s",	\
	       _pres, _phase, _stub);				\
	exit(1);						\
}

/* Used for internal flick errors */
enum {
	FLICK_STATE_NONE,
	FLICK_STATE_PROLOGUE,
	FLICK_STATE_MARSHAL,
	FLICK_STATE_SEND,
	FLICK_STATE_SEND_RECEIVE,
	FLICK_STATE_FUNCTION_CALL,
	FLICK_STATE_FUNCTION_RETURN,
	FLICK_STATE_RECEIVE,
	FLICK_STATE_UNMARSHAL,
	FLICK_STATE_EPILOGUE,
	FLICK_STATE_MAX
};

struct flick_stub_state {
	int state;
	int error_number;
};

#define FLICK_STATE_TO_SERVER_START(fstate, LINKNAME) {		\
	switch(fstate.state) {					\
	case FLICK_STATE_PROLOGUE:				\
	case FLICK_STATE_RECEIVE:				\
	case FLICK_STATE_UNMARSHAL:				\
		flick_##LINKNAME##_server_end_decode();		\
	case FLICK_STATE_FUNCTION_CALL:				\
	case FLICK_STATE_FUNCTION_RETURN:			\
	case FLICK_STATE_MARSHAL:				\
	case FLICK_STATE_SEND:					\
	case FLICK_STATE_EPILOGUE:				\
		break;						\
	}							\
	switch(fstate.state) {					\
	case FLICK_STATE_PROLOGUE:				\
	case FLICK_STATE_RECEIVE:				\
	case FLICK_STATE_UNMARSHAL:				\
	case FLICK_STATE_FUNCTION_CALL:				\
	case FLICK_STATE_FUNCTION_RETURN:			\
		flick_##LINKNAME##_server_start_encode();	\
		break;						\
	case FLICK_STATE_MARSHAL:				\
		flick_##LINKNAME##_server_restart_encode();	\
		break;						\
	case FLICK_STATE_SEND:					\
	case FLICK_STATE_EPILOGUE:				\
		break;						\
	}							\
}

#define flick_stub_error(errval, _handler) {			\
	_stub_state.error_number = errval;			\
	goto _handler;						\
}

/* This is used when no allocation/deallocation is necessary */
#define null_flick_alloc(a)
#define null_flick_free(a)

/* This is used when automatic deallocation on function exit is wanted */
#define auto_flick_alloc(a) alloca(a)
#define auto_flick_free(a)

/* This is used to allocate a handle for an ``in'' parameter.
   It is used by the decomposed presentation style to keep the
   actual pointer to the data within the message structure. */
#define msg_handle_flick_alloc(a) ({			\
	int _i = 0;					\
	void **_p = &(_msg->pvect[0]);			\
	while (*_p) {					\
		_p++;					\
		if (++_i == 7 && *_p) {			\
			_p = &(((void **) *_p)[0]);	\
			_i = 0;				\
		}					\
	}						\
	if (_i == 7) {					\
		*_p = calloc(sizeof(void *), 8);	\
		_p = &(((void **) *_p)[0]);		\
	}						\
							\
	(void *) _p;					\
})
#define msg_handle_flick_free(a)

/* These #defines must correspond to the strings found in
   `c/pbe/lib/mu_make_error.cc'.  */
#define FLICK_ERROR_NONE		0
#define FLICK_ERROR_CONSTANT		1
#define FLICK_ERROR_VIRTUAL_UNION	2
#define FLICK_ERROR_STRUCT_UNION	3
#define FLICK_ERROR_DECODE_SWITCH	4
#define FLICK_ERROR_COLLAPSED_UNION	5
#define FLICK_ERROR_VOID_UNION		6
#define FLICK_ERROR_COMMUNICATION	7
#define FLICK_ERROR_OUT_OF_BOUNDS	8
#define FLICK_ERROR_INVALID_TARGET	9
#define FLICK_ERROR_NO_MEMORY           10
#define FLICK_ERROR_MAX			11

#endif /* __flick_pres_all_h */
